/*********************************************************************
 *
 *                Microchip USB PICkit 2 Bootloader
 *
 *********************************************************************
 * FileName:        main.c
 * Dependencies:    See INCLUDES section below
 * Processor:       PIC18
 * Compiler:        C18 2.30.01+
 * Company:         Microchip Technology, Inc. / JAKA
 * Homepage:        http://kair.us/projects/pk2m_programmer/
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the Company) for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 * Author               Date        Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Rawin Rojvanit     11/19/04      Original.
 * Steven Bible      2005-01-18     Adapted to PICkit 2
 * Steven Bible      2005-04-20     Version 1 Released
 * Jaakko Kairus     19.2.2022      Adapted to PIC18F25K50 for PK2M
 ********************************************************************/

/** C O N F I G U R A T I O N   B I T S **************************************/

//#pragma config PLLDIV = 5, CPUDIV = OSC1_PLL2, USBDIV = 2                   // CONFIG1L
//#pragma config FOSC = HSPLL_HS, FCMEN = OFF, IESO = OFF                     // CONFIG1H
//#pragma config PWRT = ON, BOR = OFF, BORV = 3, VREGEN = ON                	// CONFIG2L
//#pragma config WDT = OFF, WDTPS = 32768                                     // CONFIG2H
//#pragma config MCLRE = OFF, LPT1OSC = OFF, PBADEN = OFF, CCP2MX = ON        // CONFIG3H
//#pragma config STVREN = ON, LVP = OFF, XINST = OFF, DEBUG = OFF				// CONFIG4L
//#pragma config CP0 = OFF, CP1 = OFF, CP2 = OFF, CP3 = OFF                   // CONFIG5L
//#pragma config CPB = OFF, CPD = OFF                                         // CONFIG5H
//#pragma config WRT0 = ON, WRT1 = OFF, WRT2 = OFF, WRT3 = OFF                // CONFIG6L
//#pragma config WRTB = ON, WRTC = OFF, WRTD = OFF                            // CONFIG6H
//#pragma config EBTR0 = OFF, EBTR1 = OFF, EBTR2 = OFF, EBTR3 = OFF           // CONFIG7L
//#pragma config EBTRB = OFF                                                  // CONFIG7H

// CONFIG 1L
#pragma config PLLSEL   = PLL3X     // 3X PLL multiplier selected
#pragma config CFGPLLEN = ON        // PLL turned on during execution
#pragma config CPUDIV   = NOCLKDIV  // 1:1 mode (for 48MHz CPU)
#pragma config LS48MHZ  = SYS48X8   // Clock div / 8 in Low Speed USB mode
// CONFIG 1H
#pragma config FOSC     = INTOSCIO  // HFINTOSC selected at powerup, no clock out
#pragma config PCLKEN   = OFF       // Primary oscillator driver
#pragma config FCMEN    = OFF       // Fail safe clock monitor
#pragma config IESO     = OFF       // Internal/external switchover (two speed startup)
// CONFIG 2L
#pragma config nPWRTEN  = ON        // Power up timer (typically 64.4 ms)
// #pragma config BOREN    = OFF       // BOR disabled
#pragma config BOREN    = SBORDIS   // BOR enabled in HW, control from firmware disabled
#pragma config BORV     = 285		// BOR voltage 2.85V
#pragma config nLPBOR   = OFF       // Low Power BOR
// CONFIG 2H
#pragma config WDTEN    = OFF       // Watchdog Timer disabled in hardware
#pragma config WDTPS    = 32768     // WDT postscalar
// CONFIG 3H
#pragma config MCLRE    = OFF       // MCLR function disabled (RE3 enabled)
#pragma config SDOMX    = RC7       // SDO function location
#pragma config T3CMX    = RC0		// Timer clock input location
#pragma config PBADEN   = OFF       // Port B Digital/Analog Powerup Behavior
#pragma config CCP2MX   = RC1		// CCP2 output at RC1
// CONFIG 4L
#pragma config XINST    = OFF       // Extended instruction set
#pragma config LVP      = OFF       // Low voltage programming
#pragma config STVREN   = ON        // Stack overflow reset
// CONFIG 5L
#pragma config CP0      = OFF		// Block 0 not code protected
#pragma config CP1      = OFF		// Block 1 not code protected
#pragma config CP2      = OFF		// Block 2 not code protected
#pragma config CP3      = OFF		// Block 3 not code protected
// CONFIG 5H
#pragma config CPD      = OFF		// EEPROM not code protected
#pragma config CPB      = OFF		// Boot block not code protected

#pragma config WRT0 = ON, WRT1 = OFF, WRT2 = OFF, WRT3 = OFF                // CONFIG6L
#pragma config WRTB = ON, WRTC = ON, WRTD = OFF                             // CONFIG6H
#pragma config EBTR0 = OFF, EBTR1 = OFF, EBTR2 = OFF, EBTR3 = OFF           // CONFIG7L
#pragma config EBTRB = OFF                                                  // CONFIG7H


/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include "system\typedefs.h"                        // Required
#include "system\usb\usb.h"                         // Required
#include "io_cfg.h"                                 // Required
#include "user\boot.h"
#include <delays.h>

#include "system\usb\usb_compile_time_validation.h" // Optional


/** V A R I A B L E S ********************************************************/

#pragma udata


/** P R I V A T E  P R O T O T Y P E S ***************************************/

static void InitializeSystem(void);
void USBTasks(void);
void main_loop(void);


/** V E C T O R  R E M A P P I N G *******************************************/

#pragma code _HIGH_INTERRUPT_VECTOR = 0x000008
void _high_ISR (void)
{
    _asm goto RM_HIGH_INTERRUPT_VECTOR _endasm
}

#pragma code _LOW_INTERRUPT_VECTOR = 0x000018
void _low_ISR (void)
{
    _asm goto RM_LOW_INTERRUPT_VECTOR _endasm
}


/** D E C L A R A T I O N S **************************************************/

#pragma code


/******************************************************************************
 * Function:        void main (void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Main program entry point.
 *
 * Note:            None
 *****************************************************************************/
void main (void) {

    byte Temp;
//    word Counter;

	//----------------------------------------
	// Ensure UART pins are tri-stated
	//----------------------------------------
	tris_UART_TX = 1;
	tris_UART_RX = 1;

    //----------------------------------------
    // Check Bootload Mode Entry Condition
    //----------------------------------------

	ANSELB = 0x00;					// Set all port B pins to digital
    tris_PROG_SWITCH = 1;           // RB5 Input
    // PROG_SWITCH_pin = 1;                // initialize port to 1 (why?)
    tris_Vdd_TGT_P = 0;				// RB4 Output
    Vdd_TGT_P = 1;					// Disable target power (prevents dim glow of TARGET LED)
    INTCON2bits.NOT_RBPU = 0;       // initialize PORTB pull-ups on
	WPUB = 0xff;					// Enable all PORTB pull-ups

    // Delay10KTCYx(600);              // delay ~500 ms (orig. delay setting)
	//BUSY_LED = 1;                           // Busy LED On
    //tris_BUSY_LED = 0;                      // Output
	Delay10KTCYx(6);              // delay ~240 ms on 25K50 w/ int osc.
	//BUSY_LED = 0;                           // Busy LED Off


    if (PROG_SWITCH_pin) {         // is the push button pressed?

        // if no, does last program memory location contain 0x55?

#if defined(__18F24K50)
        Temp = * ((rom far char *) 0x3FFE);
#else
        Temp = * ((rom far char *) 0x7FFE);
#endif
        if(Temp == 0x55) {

            _asm goto RM_RESET_VECTOR _endasm

        }

    } // end if (PROG_SWITCH)

    //----------------------------------------
    // Bootloader Mode
    //----------------------------------------
	main_loop();
	/*
    InitializeSystem();

    while(1) {

        USBDriverService();                 // Interrupt or polling method

        BootService();              // See boot.c

        if (Counter == 0)           // blink Busy LED
            BUSY_LED ^= 1;

        Counter--;

    }
    */

} // end main

void main_loop(void) {
    word Counter;

	InitializeSystem();

    while(1) {

        USBDriverService();                 // Interrupt or polling method

        BootService();              // See boot.c

        if (Counter == 0)           // blink Busy LED
            BUSY_LED ^= 1;

        Counter--;

    }
}    

/******************************************************************************
 * Function:        static void InitializeSystem(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        InitializeSystem is a centralize initialization routine.
 *                  All required USB initialization routines are called from
 *                  here.
 *
 *                  User application initialization routine should also be
 *                  called from here.                  
 *
 * Note:            None
 *****************************************************************************/
static void InitializeSystem(void)
{
    byte i;
    //Configure oscillator settings for clock settings compatible with USB 
    //operation.  Note: Proper settings depends on USB speed (full or low).
    OSCTUNE = 0x80; //3X PLL ratio mode selected
    OSCCON = 0x70;  //Switch to 16MHz HFINTOSC
    OSCCON2 = 0x10; //Enable PLL, SOSC, PRI OSC drivers turned off
    while(OSCCON2bits.PLLRDY != 1);   //Wait for PLL lock
	ACTCON = 0x90;  //Enable active clock tuning for USB operation

    #if defined(USE_USB_BUS_SENSE_IO)       // see usbcfg.h
        tris_usb_bus_sense = INPUT_PIN;     // See io_cfg.h
    #endif
    
    #if defined(USE_SELF_POWER_SENSE_IO)    // see usbcfg.h
        tris_self_power = INPUT_PIN;
    #endif

    for (i=0; i<HID_INT_OUT_EP_SIZE; i++) {
	    hid_report_out[i] = 0;
	}
 
    mInitializeUSBDriver();                 // See usbdrv.h

    // Init UnitID buffer for USB
    // no Unit ID default
    i=2; // 4 bytes
    unit_id_usb_buf[1] = DSC_STR;   // USB descriptor string
    unit_id_usb_buf[2] = 0x09;
    unit_id_usb_buf[3] = 0x04;

    /*
    if (EE_ReadByte(UNIT_ID) == '#')
    {
        for (; i < (2 * UNITID_MAX_LEN); i+=2)
        {
            unit_id_usb_buf[i] = EE_ReadByte(UNIT_ID + (i/2));
            unit_id_usb_buf[i + 1] = 0;
            if (unit_id_usb_buf[i] == 0)
                break;
        }
    }
    */
    unit_id_usb_buf[0] = i + 2;


    BUSY_LED = 0;                           // Busy LED Off
    tris_BUSY_LED = 0;                      // Output

	USBCheckBusStatus();

} // end InitializeSystem

#pragma code jumptoloader = 0x0FFC
static void jumptoloader(void) {
	main_loop();
}


#pragma code user = RM_RESET_VECTOR


/** EOF main.c ***************************************************************/
